// -*- C++ -*-


/// \addtogroup device
/// \{

#pragma once


#include "goopax_license.h"
#include "memory.h"
#include "device.h"

namespace goopax
{

  namespace impl
  {
      namespace interface
    {
      extern "C"
      {
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) size_t goopax_num_devices(envmode, Uint, const char*, const Uint8_t*) noexcept;
          GOOPAX_DLL_EXPORT(3.0.3) size_t goopax_num_devices2(envmode, Uint, void(*)(), const char*, const Uint8_t*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_get_devices(goopax_device*, envmode, Uint, const char*, const Uint8_t*) noexcept;
          GOOPAX_DLL_EXPORT(3.0.3) void goopax_get_devices2(goopax_device*, envmode, Uint, void(*)(), const char*, const Uint8_t*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) goopax_device_impl* goopax_get_default_device(envmode, Uint, const char*, const Uint8_t*) noexcept;
          GOOPAX_DLL_EXPORT(3.0.3) goopax_device_impl* goopax_get_default_device2(envmode, Uint, void(*)(), const char*, const Uint8_t*) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_init_params(Int*, char***, const char*) noexcept;
	GOOPAX_DLL_EXPORT(3.1.0) void goopax_register_garbage(void*, void(*)(void*)) noexcept;
      }
    }

    template <typename T> static void garbage_deleter(void* p)
    {
      delete static_cast<T*>(p);
    }
    

    static const Uint8_t sig[256] = {GOOPAX_CERT_SIG};
    
    inline vector<goopax_device> devices(envmode backend = env_GPU)
    {
        /// \ingroup device
        /// \brief Get vector of available devices for the given backend(s)
        /// Devices that are both CUDA and OpenCL capable will show up both for the env_CUDA and for the env_CL backend,
        /// but they will only be returned once if the both env_CL and env_CUDA flags are selected. Equally, devices that support the Metal backend may also show up if the env_CL backend is selected without the env_METAL backend.
        /// \return vector of devices
        /// \param backend which backend(s) to use.
      vector<goopax_device> devices(interface::goopax_num_devices2(backend, Uint(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig));
      interface::goopax_get_devices2(devices.data(), backend, Uint(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
      return devices;
    }
      
    inline goopax_device default_device(envmode backend = env_GPU)
    {
        /// \ingroup device
        /// \brief Get the fastest available device for the given backend(s)
        /// \return vector of devices
        /// \param backend which backend(s) to use.
      return interface::goopax_get_default_device2(backend, Uint(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
    }

    inline void init_params(Int& argc, char**& argv, const char* paramfile=nullptr)
    {
        /// \ingroup device
        /// May be called to allow passing command line arguments to GOOPAX or to PARAM or PARAMOPT user values.
        /// \param argc argc from the main() function
        /// \param argv argv from the main() function
        /// \param paramfile optional filename of configuration file
      check_errno(interface::goopax_E_init_params(&argc, &argv, paramfile));
    }
    inline void init_params(const char* paramfile)
    {
      check_errno(interface::goopax_E_init_params(nullptr, nullptr, paramfile));
    }
  }

  using impl::devices;
  using impl::default_device;
  using impl::goopax_device;
  using impl::init_params;
}

/// \}
