// -*- C++ -*-

#pragma once

#include "kernel.h"
#include "gpu_types.h"
#include "gpu-utils.h"


///
/// \addtogroup math
/// \{

#ifdef __CYGWIN__
namespace std
{
  using ::cbrt;
}
#endif



namespace goopax
{
  namespace impl
  {
    namespace interface
    {
      extern "C"
      {
#define MAKEFUNC1(FUNC)							\
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_math_ ## FUNC(gpu_writer_base_impl* r, gpu_writer_base_impl* a, Int MODE) noexcept;
	GOOPAX_INSTANTIATE_MATH(MAKEFUNC1)

#undef MAKEFUNC1
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_math_root(gpu_writer_base_impl* r, gpu_writer_base_impl* a, Int MODE) noexcept;


	
	GOOPAX_DLL_EXPORT(3.2.0) errno_type goopax_E_math(gpu_writer_base_impl* r, gpu_writer_base_impl* a, const char* FUNCNAME) noexcept;
	

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_math_pow_int_uint_floating_point(gpu_writer_base_impl*, gpu_writer_base_impl*, Int, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.2.0) errno_type goopax_E_math_pow(gpu_writer_base_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
      }
    }

#define MAKEFUNC(FUNC)				\
      template <class T, class TA> inline auto FUNC(checknan<T, TA> a) -> decltype(FUNC(unrange(a))) \
      { return FUNC(unrange(a)); }

    GOOPAX_INSTANTIATE_MATH_ARG(MAKEFUNC)

      
#undef MAKEFUNC


      
      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type>
      inline gpu_type<T> pow(gpu_type<T> a, Int ENUM, Uint DENOM=1)
      {
	/// Returns a to the power (ENUM/DENOM). \ingroup math
	gpu_type<T> ret;
	check_errno(interface::goopax_E_math_pow_int_uint_floating_point(ret.impl.get(), a.impl.get(), ENUM, DENOM));
	return ret;
      }

      template <Int ENUM, Uint DENOM, class T, class ENABLE=typename enable_if<is_floating_point<T>::value>::type>
      gpu_type<T> pow(const gpu_type<T> a)
      {
	/// Returns a to the power (ENUM/DENOM). \ingroup math
	return pow(a, ENUM, DENOM);
      }

      template <class T, class ENABLE=typename enable_if<is_floating_point<T>::value>::type>
      gpu_type<T> pow(const gpu_type<T> a, const gpu_type<T> b)
      {
	/// Returns a to the power b. \ingroup math
	gpu_type<T> ret;
	check_errno(interface::goopax_E_math_pow(ret.impl.get(), a.impl.get(), b.impl.get()));
	return ret;
      }
      
      template <class T, class D, class ENABLE=typename enable_if<is_floating_point<T>::value && is_floating_point<D>::value>::type>
					    gpu_type<T> pow(const gpu_type<T> a, const D b)
      {
	/// Returns a to the power b. \ingroup math
	return pow(a, gpu_type<T>(b));
      }
      
      /// Square root \ingroup math
      inline gpu_half sqrt(gpu_half a) { return pow<1,2>(a); }
      /// Square root \ingroup math
      inline gpu_float sqrt(gpu_float a) { return pow<1,2>(a); }
      /// Square root \ingroup math
      inline gpu_double sqrt(gpu_double a) { return pow<1,2>(a); }
      /// Cubic root \ingroup math
      inline gpu_half cbrt(gpu_half a) { return pow<1,3>(a); }
      /// Cubic root \ingroup math
      inline gpu_float cbrt(gpu_float a) { return pow<1,3>(a); }
      /// Cubic root \ingroup math
      inline gpu_double cbrt(gpu_double a) { return pow<1,3>(a); }
    /*
      /// returns e^a \ingroup math
      inline gpu_half exp(gpu_half a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns e^a \ingroup math
      inline gpu_float exp(gpu_float a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns e^a \ingroup math
      inline gpu_double exp(gpu_double a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns log_n(a) \ingroup math
      inline gpu_half log(gpu_half a) { return log2(a) * impl::log(2.0); }
      /// returns log_n(a) \ingroup math
      inline gpu_float log(gpu_float a) { return log2(a) * impl::log(2.0); }
      /// returns log_n(a) \ingroup math
      inline gpu_double log(gpu_double a) { return log2(a) * impl::log(2.0); }
      /// sin \ingroup math
      inline gpu_half sin(gpu_half a) { return sinpi(a * (1.0 / PI)); }
      /// sin \ingroup math
      inline gpu_float sin(gpu_float a) { return sinpi(a * (1.0 / PI)); }
      /// sin \ingroup math
      inline gpu_double sin(gpu_double a) { return sinpi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_half cos(gpu_half a) { return cospi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_float cos(gpu_float a) { return cospi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_double cos(gpu_double a) { return cospi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_half tan(gpu_half a) { return tanpi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_float tan(gpu_float a) { return tanpi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_double tan(gpu_double a) { return tanpi(a * (1.0 / PI)); }
      /// asin \ingroup math
      inline gpu_half asin(gpu_half a) { return asinpi(a) * PI; }
      /// asin \ingroup math
      inline gpu_float asin(gpu_float a) { return asinpi(a) * PI; }
      /// asin \ingroup math
      inline gpu_double asin(gpu_double a) { return asinpi(a) * PI; }
      /// acos \ingroup math
      inline gpu_half acos(gpu_half a) { return acospi(a) * PI; }
      /// acos \ingroup math
      inline gpu_float acos(gpu_float a) { return acospi(a) * PI; }
      /// acos \ingroup math
      inline gpu_double acos(gpu_double a) { return acospi(a) * PI; }
	/// atan \ingroup math
	inline gpu_half atan(gpu_half a) { return atanpi(a) * PI; }
	/// atan \ingroup math
	inline gpu_float atan(gpu_float a) { return atanpi(a) * PI; }
	/// atan \ingroup math
	inline gpu_double atan(gpu_double a) { return atanpi(a) * PI; }
    */
    

    template <typename gpu_T> gpu_T mathfunc(gpu_T a, const char* funcname)
    {
      gpu_T ret;
      check_errno(interface::goopax_E_math(ret.impl.get(), a.impl.get(), funcname));
      return ret;
    }
    
    inline gpu_half sin(gpu_half a) { return mathfunc(a, "sin"); }
    inline gpu_float sin(gpu_float a) { return mathfunc(a, "sin"); }
    inline gpu_double sin(gpu_double a) { return mathfunc(a, "sin"); }
    inline gpu_half sinpi(gpu_half a) { return mathfunc(a, "sinpi"); }
    inline gpu_float sinpi(gpu_float a) { return mathfunc(a, "sinpi"); }
    inline gpu_double sinpi(gpu_double a) { return mathfunc(a, "sinpi"); }
    inline gpu_half cos(gpu_half a) { return mathfunc(a, "cos"); }
    inline gpu_float cos(gpu_float a) { return mathfunc(a, "cos"); }
    inline gpu_double cos(gpu_double a) { return mathfunc(a, "cos"); }
    inline gpu_half cospi(gpu_half a) { return mathfunc(a, "cospi"); }
    inline gpu_float cospi(gpu_float a) { return mathfunc(a, "cospi"); }
    inline gpu_double cospi(gpu_double a) { return mathfunc(a, "cospi"); }
    inline gpu_half tan(gpu_half a) { return mathfunc(a, "tan"); }
    inline gpu_float tan(gpu_float a) { return mathfunc(a, "tan"); }
    inline gpu_double tan(gpu_double a) { return mathfunc(a, "tan"); }
    inline gpu_half tanpi(gpu_half a) { return mathfunc(a, "tanpi"); }
    inline gpu_float tanpi(gpu_float a) { return mathfunc(a, "tanpi"); }
    inline gpu_double tanpi(gpu_double a) { return mathfunc(a, "tanpi"); }
    inline gpu_half exp(gpu_half a) { return mathfunc(a, "exp"); }
    inline gpu_float exp(gpu_float a) { return mathfunc(a, "exp"); }
    inline gpu_double exp(gpu_double a) { return mathfunc(a, "exp"); }
    inline gpu_half exp2(gpu_half a) { return mathfunc(a, "exp2"); }
    inline gpu_float exp2(gpu_float a) { return mathfunc(a, "exp2"); }
    inline gpu_double exp2(gpu_double a) { return mathfunc(a, "exp2"); }
    inline gpu_half log(gpu_half a) { return mathfunc(a, "log"); }
    inline gpu_float log(gpu_float a) { return mathfunc(a, "log"); }
    inline gpu_double log(gpu_double a) { return mathfunc(a, "log"); }
    inline gpu_half log2(gpu_half a) { return mathfunc(a, "log2"); }
    inline gpu_float log2(gpu_float a) { return mathfunc(a, "log2"); }
    inline gpu_double log2(gpu_double a) { return mathfunc(a, "log2"); }
    inline gpu_half asin(gpu_half a) { return mathfunc(a, "asin"); }
    inline gpu_float asin(gpu_float a) { return mathfunc(a, "asin"); }
    inline gpu_double asin(gpu_double a) { return mathfunc(a, "asin"); }
    inline gpu_half asinpi(gpu_half a) { return mathfunc(a, "asinpi"); }
    inline gpu_float asinpi(gpu_float a) { return mathfunc(a, "asinpi"); }
    inline gpu_double asinpi(gpu_double a) { return mathfunc(a, "asinpi"); }
    inline gpu_half acos(gpu_half a) { return mathfunc(a, "acos"); }
    inline gpu_float acos(gpu_float a) { return mathfunc(a, "acos"); }
    inline gpu_double acos(gpu_double a) { return mathfunc(a, "acos"); }
    inline gpu_half acospi(gpu_half a) { return mathfunc(a, "acospi"); }
    inline gpu_float acospi(gpu_float a) { return mathfunc(a, "acospi"); }
    inline gpu_double acospi(gpu_double a) { return mathfunc(a, "acospi"); }
    inline gpu_half atan(gpu_half a) { return mathfunc(a, "atan"); }
    inline gpu_float atan(gpu_float a) { return mathfunc(a, "atan"); }
    inline gpu_double atan(gpu_double a) { return mathfunc(a, "atan"); }
    inline gpu_half atanpi(gpu_half a) { return mathfunc(a, "atanpi"); }
    inline gpu_float atanpi(gpu_float a) { return mathfunc(a, "atanpi"); }
    inline gpu_double atanpi(gpu_double a) { return mathfunc(a, "atanpi"); }
    inline gpu_half sinh(gpu_half a) { return mathfunc(a, "sinh"); }
    inline gpu_float sinh(gpu_float a) { return mathfunc(a, "sinh"); }
    inline gpu_double sinh(gpu_double a) { return mathfunc(a, "sinh"); }
    inline gpu_half cosh(gpu_half a) { return mathfunc(a, "cosh"); }
    inline gpu_float cosh(gpu_float a) { return mathfunc(a, "cosh"); }
    inline gpu_double cosh(gpu_double a) { return mathfunc(a, "cosh"); }
    inline gpu_half tanh(gpu_half a) { return mathfunc(a, "tanh"); }
    inline gpu_float tanh(gpu_float a) { return mathfunc(a, "tanh"); }
    inline gpu_double tanh(gpu_double a) { return mathfunc(a, "tanh"); }
    inline gpu_half asinh(gpu_half a) { return mathfunc(a, "asinh"); }
    inline gpu_float asinh(gpu_float a) { return mathfunc(a, "asinh"); }
    inline gpu_double asinh(gpu_double a) { return mathfunc(a, "asinh"); }
    inline gpu_half acosh(gpu_half a) { return mathfunc(a, "acosh"); }
    inline gpu_float acosh(gpu_float a) { return mathfunc(a, "acosh"); }
    inline gpu_double acosh(gpu_double a) { return mathfunc(a, "acosh"); }
    inline gpu_half atanh(gpu_half a) { return mathfunc(a, "atanh"); }
    inline gpu_float atanh(gpu_float a) { return mathfunc(a, "atanh"); }
    inline gpu_double atanh(gpu_double a) { return mathfunc(a, "atanh"); }
    inline gpu_half erf(gpu_half a) { return mathfunc(a, "erf"); }
    inline gpu_float erf(gpu_float a) { return mathfunc(a, "erf"); }
    inline gpu_double erf(gpu_double a) { return mathfunc(a, "erf"); }
    inline gpu_half erfc(gpu_half a) { return mathfunc(a, "erfc"); }
    inline gpu_float erfc(gpu_float a) { return mathfunc(a, "erfc"); }
    inline gpu_double erfc(gpu_double a) { return mathfunc(a, "erfc"); }
    inline gpu_half tgamma(gpu_half a) { return mathfunc(a, "tgamma"); }
    inline gpu_float tgamma(gpu_float a) { return mathfunc(a, "tgamma"); }
    inline gpu_double tgamma(gpu_double a) { return mathfunc(a, "tgamma"); }

  }
}
/// \}
