// -*- C++ -*-


#pragma once

#include "../proto.h"
#include "./defs.h"
#include "stdinc.h"
#include "vec.h"
#include <cstdint>

#if defined(GOOPAX_USE_BOOST) && GOOPAX_USE_BOOST
#ifdef _WIN32
#include <windows.h>
#include "clean_environment.h"
#endif
#include <boost/thread.hpp>
#else
#include <mutex>
#endif

#include "exception.h"
#include "stdinc.h"
#include "operators.h"



namespace goopax
{
  namespace impl
  {
    template <typename T> struct is_gpu_struct;
  }
}


namespace goopax
{
  namespace impl
  {
    template <typename STR> struct output_goopax_struct_helper;
    template <typename T> output_goopax_struct_helper<T> output_goopax_struct(const T& t);
  }
}


namespace goopax
{
  template <typename T, typename ENABLE=void> struct goopax_struct_type;
  template <class STR, class X, class ENABLE=void> struct goopax_struct_changetype;

}


namespace goopax
{
  namespace impl
  {
#if defined(GOOPAX_USE_BOOST) && GOOPAX_USE_BOOST
    using boost::mutex;
    using boost::lock_guard;
    using boost::unique_lock;
#else
    using std::mutex;
    using std::lock_guard;
    using std::unique_lock;
#endif

  }

  namespace impl
  {
    enum init_state
      {
	init_yes = 12345,
	init_no = 99999,
	init_destructed = 55555
      };



    namespace interface
    {
#if __cplusplus >= 201700
      using demangle_deletefunc_t = void(const char*) noexcept;
#else
	using demangle_deletefunc_t = void(const char*);
#endif
      extern "C"
      {
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) const char* goopax_type_demangle_gnuc2(const char*) noexcept;
	GOOPAX_DLL_EXPORT(3.1.0) const char* goopax_type_demangle_gnuc3(const char*, demangle_deletefunc_t**) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_range_throw_failure2(init_state init, const char* tiname) noexcept;
      }
    }


    inline string pretty_typename(const char* tiname)
    {
      interface::demangle_deletefunc_t* deleter;
      const char* tmp = interface::goopax_type_demangle_gnuc3(tiname, &deleter);
      string ret;
      if (tmp == nullptr)
	{
	  ret = "[unknown]";
	}
      else
	{
	  ret = tmp;
	  deleter(tmp);
	}
      return ret;
    }

    inline string pretty_typename(const type_info& ti)
    {
      return pretty_typename(ti.name());
    }




    template <typename T> struct is_checknan :
      public false_type
    {};

    template <typename T, typename THREAD_ACCESS> struct is_checknan<checknan<T, THREAD_ACCESS>> :
      public true_type
    {};


    template <typename T> struct unrangetype_core
    {
      using type = T;
    };
    template <typename T, typename TA> struct unrangetype_core<checknan<T, TA> >
    {
      typedef T type;
    };


    
    template <typename T> struct unrangetype;

    template <typename T, typename=void> struct unrangetype_help;
    /*
    template <typename T, typename=void> struct unrangetype_help
    {
      using type = T;
    };
    */
    template <typename T> struct unrangetype_help<T, typename enable_if<is_arithmetic<T>::value || is_pointer<T>::value>::type>
    {
      using type = T;
    };
    
    template <typename T> struct unrangetype_help<T, typename enable_if<is_container<T>::value && !is_derived_from_array<T>::value>::type>
    {
      using type = typename changetype<T, typename unrangetype<typename T::value_type>::type>::type;
    };
    
    template <typename T> struct unrangetype :
      public unrangetype_help<T>
    {
    };

    template <typename T, typename TA> struct unrangetype<checknan<T, TA> >
    {
      typedef T type;
    };
    template <typename T> struct unrangetype<goopax::impl::gpu_type<T>>
    {
      using type = goopax::impl::gpu_type<T>;
    };
    
    template <typename T, size_t N> struct unrangetype<std::array<T, N>>
    {
      using type = std::array<typename unrangetype<T>::type, N>;
    };
#if GOOPAX_HAVE_DEBUG_STL
    template <typename T, size_t N> struct unrangetype<std::__debug::array<T, N>>
    {
      using type = std::__debug::array<typename unrangetype<T>::type, N>;
    };
#endif
    
    template <typename A, typename B> struct unrangetype<pair<A, B> >
    {
      using type = pair<typename unrangetype<A>::type, typename unrangetype<B>::type>;
    };


      template <typename T, size_t N> inline std::array<typename unrangetype<T>::type, N> unrange(const std::array<T, N>& a);
#if GOOPAX_HAVE_DEBUG_STL
      template <typename T, size_t N> inline std::__debug::array<typename unrangetype<T>::type, N> unrange(const std::__debug::array<T, N>& a);
#endif
      template <typename T> inline
      typename conditional<is_base_of<typename unrangetype<T>::type, T>::value, const typename unrangetype<T>::type&, typename unrangetype<T>::type>::type
      unrange(const T& a);

      
      template <typename T, size_t N> inline std::array<typename unrangetype<T>::type, N> unrange(const std::array<T, N>& a)
    {
      std::array<typename unrangetype<T>::type, N> ret;
      for (size_t k=0; k<N; ++k)
	ret[k] = unrange(a[k]);
      return ret;
    }
#if GOOPAX_HAVE_DEBUG_STL
    template <typename T, size_t N> inline std::__debug::array<typename unrangetype<T>::type, N> unrange(const std::__debug::array<T, N>& a)
    {
      std::__debug::array<typename unrangetype<T>::type, N> ret;
      for (size_t k=0; k<N; ++k)
	ret[k] = unrange(a[k]);
      return ret;
    }
#endif
    
    template <typename T> inline 
    typename conditional<is_base_of<typename unrangetype<T>::type, T>::value, const typename unrangetype<T>::type&, typename unrangetype<T>::type>::type 
    unrange(const T& a)
    {
      return a;
    }
    
#if GOOPAX_DEBUG_FOR_CL
    template <typename T> struct cldebug_unrangetype :
      public unrangetype<T>
    {};
#else
    template <typename T> struct cldebug_unrangetype
    {
      typedef T type;
    };
#endif


    template <typename T, typename=void> struct is_single_number  :
      false_type
    {};

    template <typename T> struct is_single_number<T,
						  typename enable_if<is_fundamental<typename unrangetype_core<T>::type>::value || is_same<typename unrangetype<T>::type, half>::value>::type> :
      true_type
    {};
    
    template <typename T> struct is_single_number<goopax::impl::gpu_type<T>> :
      true_type
    {};
    


    
    template <typename ID_t> struct ta_mutex_base
    {
      mutable ID_t read_thread;   // -2: several, -1: none, 0..i..N thread i
      mutable ID_t write_thread;
      mutable uint32_t accesstime;
      mutable bool atomic_access;
    };

    

    namespace interface
    {
      template <typename=void> struct callmap
      {
	static callmap data;
	void(*barrier)(int, unsigned int, unsigned int, void*) noexcept = nullptr;
	void(*ta_ra_local)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_wa_local)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_ra_atomic_local)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_wa_atomic_local)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_ra_wf)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_wa_wf)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_ra_atomic_wf)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;
	void(*ta_wa_atomic_wf)(const ta_mutex_base<uint16_t>*) noexcept = nullptr;

	void(*range_throw_failure)(init_state init, const char* ti) noexcept = nullptr;
	uint32_t(*ta_local_init)() noexcept = nullptr;
	uint32_t(*ta_wf_init)() noexcept = nullptr;

	void(*ta_inc_fbt)() noexcept = nullptr;
	void(*threadcall) (void(*)(unsigned int, unsigned int, void* const* const, void*), unsigned int, unsigned int, void* const* const) noexcept= nullptr;
	void(*dumpstream_print)(const checknan<unsigned int>*, const unsigned int) noexcept = nullptr;
	void(*io_warn)(const char*, unsigned int) noexcept = nullptr;
      };
      template <typename T> callmap<T> callmap<T>::data;      
    }

    struct warning
    {
      unique_ptr<stringstream> ss;
      unsigned int MODE = 0;
      
      warning() :
	ss(new stringstream())
      {}
      warning& operator = (warning&&) = delete;
      inline ~warning()
      {
	if (ss.get() != nullptr)
	  {
	    const string sss = ss->str();
#ifdef GOOPAX_CPU_CALL
	    interface::callmap<>::data.io_warn(sss.c_str(), MODE);
#else
	    interface::goopax_io_warn3(sss.c_str(), MODE);
#endif
	  }
      }
      
      template <class T> friend warning&& operator << (warning&& s, const T& t)
      {
	using std::operator <<;
	static_cast<ostream&>(*s.ss) << t;
	return move(s);
      }
      friend warning&& operator << (warning&& s, ostream& (*func)(ostream&))
      {
	using std::operator <<;
	static_cast<ostream&>(*s.ss) << func;
	return move(s);
      }
    };

    struct exwhat :
      public warning
    {
      exwhat()
      {
	this->MODE = 1;
      }
    };
    
    struct thread_access_none
    {
      inline void write_access() {}
      inline void read_access() const {}
      inline void atomic_write_access() {}
      inline void atomic_read_access() const {}
      inline thread_access_none() {}
      inline ~thread_access_none() {}
    };


    struct ta_mutex_local :
      public ta_mutex_base<uint16_t>
    {
      inline void read_access() const
      {
	interface::callmap<>::data.ta_ra_local(this);
      }
      inline void write_access()
      {
	interface::callmap<>::data.ta_wa_local(this);
      }
      inline void atomic_read_access() const
      {
	interface::callmap<>::data.ta_ra_atomic_local(this);
      }
      inline void atomic_write_access()
      {
	interface::callmap<>::data.ta_wa_atomic_local(this);
      }
      inline ta_mutex_local()
      {
	this->accesstime = interface::callmap<>::data.ta_local_init();
      }
    };

    struct ta_mutex_wf :
      public ta_mutex_base<uint16_t>
    {
      inline void read_access() const
      {
	interface::callmap<>::data.ta_ra_wf(this);
      }
      inline void write_access()
      {
	interface::callmap<>::data.ta_wa_wf(this);
      }
      inline void atomic_read_access() const
      {
	interface::callmap<>::data.ta_ra_atomic_wf(this);
      }
      inline void atomic_write_access()
      {
	interface::callmap<>::data.ta_wa_atomic_wf(this);
      }
      inline ta_mutex_wf()
      {
	this->accesstime = interface::callmap<>::data.ta_wf_init();
      }
    };

    struct thread_access_local
    {
      mutable mutex M;
      
      ta_mutex_local sub_thread;
      void read_access() const
      {
	lock_guard<mutex> lock(M);
	sub_thread.read_access();
      }
      void write_access()
      {
	lock_guard<mutex> lock(M);
	sub_thread.write_access();
      }
      void atomic_read_access() const
      {
	sub_thread.atomic_read_access();
      }
      void atomic_write_access()
      {
	sub_thread.atomic_write_access();
      }

      inline thread_access_local() = default;
      
      thread_access_local(const thread_access_local&) = delete;
      thread_access_local& operator = (const thread_access_local&) = delete;
    };

    struct thread_access_full :
      private thread_access_local
    {
      using thread_access_local::M;
      ta_mutex_wf sub_wf;
      void read_access() const
      {
	lock_guard<mutex> lock(M);
	sub_wf.read_access();
	sub_thread.read_access();
      }
      void write_access()
      {
	lock_guard<mutex> lock(M);
	sub_wf.write_access();
	sub_thread.write_access();
      }
      void atomic_read_access() const
      {
	sub_wf.atomic_read_access();
	sub_thread.atomic_read_access();
      }
      void atomic_write_access()
      {
	sub_wf.atomic_write_access();
	sub_thread.atomic_write_access();
      }

      inline thread_access_full() = default;
      
      thread_access_full(const thread_access_full&) = delete;
      thread_access_full& operator = (const thread_access_full&) = delete;
    };



    struct checknan_base
    {
      volatile init_state init;

      GOOPAX_NORETURN inline void throw_failure(const type_info& ti) const
      {
#ifdef GOOPAX_CPU_CALL
	interface::callmap<>::data.range_throw_failure(init, ti.name());
#else
	interface::goopax_range_throw_failure2(init, ti.name());
#endif
	abort();
      }

      inline checknan_base() :
	init(init_no)
      {}
  
      inline ~checknan_base()
      {
	init = init_destructed;
      }
      checknan_base(const checknan_base&) = default;
      checknan_base& operator = (const checknan_base&) = default;
    };


    template <typename T, typename THREAD_ACCESS> class checknan :
      private checknan_base
    {
        template <typename U, typename UA> friend class checknan;
        
      static_assert(!is_checknan<T>::value, "");
      T data;
      THREAD_ACCESS TA;
        
        inline void check() const
        {
            if (init != init_yes)
            {
                this->throw_failure(typeid(T));
            }
        }
        void assigncheck() const
        {
            if (init != init_yes && init != init_no)
            {
                this->throw_failure(typeid(T));
            }
        }
        
    public:

    mutex& get_mutex() const
        {
            return TA.M;
        }
        
        inline bool is_uninitialized() const
        {
            return (init == init_no);
        }
        inline bool is_initialized() const
        {
            return (init == init_yes);
        }

        inline T& unrange()
        {
            check();
            TA.read_access();
            return data;
        }

        inline operator const T&() const
      {
	check();
	TA.read_access();
	return data; 
      }

      inline checknan& operator ++ ()
      {
	*this = *this + 1;
	return *this;
      }

      inline checknan& operator -- ()
      {
	*this = *this - 1;
	return *this;
      }

      inline checknan operator ++ (int)
      {
	const checknan oldval = *this;
	++*this;
	return oldval;
      }

      inline checknan operator -- (int)
      {
	const checknan oldval = *this;
	--*this;
	return oldval;
      }


      template <typename DUMMY=int> inline checknan& operator |= (const checknan& b)
      {
	*this = *this | goopax::impl::unrange(b);
	return *this;
      }

      template <typename DUMMY=int> inline checknan& operator &= (const checknan& b)
      {
	*this = *this & goopax::impl::unrange(b);
	return *this;
      }

      template <typename DUMMY=int> inline checknan& operator ^= (const checknan& b)
      {
	*this = *this ^ goopax::impl::unrange(b);
	return *this;
      }

      template <typename DUMMY=int> inline checknan& operator >>= (const unsigned int b)
      {
	*this = *this >> b;
	return *this;
      }

      template <typename DUMMY=int> inline checknan& operator <<= (const unsigned int b)
      {
	*this = *this << b;
	return *this;
      }


      checknan<T, thread_access_none> atomic_read()
      {
	checknan<T, thread_access_none> ret;
	static_cast<checknan_base&>(ret) = *this;
	if (init == init_yes)
	  ret.data = data;
	TA.atomic_read_access();
	return ret;
      }

      void atomic_write(checknan<T, thread_access_none> a)
      {
	static_cast<checknan_base&>(*this) = a;
	data = (a.init == init_yes ? (T)a.data : T());
	TA.atomic_write_access();
      }
      

      template <typename U, typename UT> inline checknan& operator = (const checknan<U, UT>& in)
      {
	in.assigncheck();
          assigncheck();
          static_cast<checknan_base&>(*this) = in;
	data = (in.init == init_yes ? (T)in.data : T());
	in.TA.read_access();
	TA.write_access();
	return *this;
      }
      inline checknan& operator = (const checknan& in)
      {
	in.assigncheck();
          assigncheck();
	static_cast<checknan_base&>(*this) = in;
	data = (in.init == init_yes ? (T)in.data : T());
	in.TA.read_access();
	TA.write_access();
	return *this;
      }
  

      template <typename U, typename UT> inline checknan(const checknan<U, UT>& in) :
	checknan_base(in),
	data(in.init == init_yes ? (T)in.data : T())
      {
	in.assigncheck();
	in.TA.read_access();
	TA.write_access();
      }

      inline checknan(const checknan& in) :
	checknan_base(in),
	data(in.init == init_yes ? (T)in.data : T())
      {
	in.assigncheck();
	in.TA.read_access();
	TA.write_access();
      }

      inline checknan(T in) :
	data(in)
      {
	this->init = init_yes;
	check();
	TA.write_access();
      }

      template <typename U=void, typename=typename enable_if<is_same<T, half>::value, U>::type>
      inline checknan(half_float::detail::expr a) :
	checknan((half)a)
      {}
      
      inline checknan()
      {}

      template <typename U> checknan& operator = (const U& in)
	{
            assigncheck();
            data = ::goopax::impl::unrange(in);
	  init = init_yes;
	  TA.write_access();
	  return *this;
	}


      friend ostream& operator << (ostream& s, const checknan& n)
      {
	if (n.init == init_yes)
	  s << n.data;
	else if (n.init == init_no)
	  s << "UNINITIALIZED";
	else if (n.init == init_destructed)
	  s << "DESTRUCTED";
	else
	  s << "UNCONSTRUCTED";
	return s;
      }
      /*
      friend ostream& operator << (stringstream& s, const checknan& n)
      {
	return static_cast<ostream&>(s) << n;
      }
      */

      friend istream& operator >> (istream& s, checknan& n)
      {
	T ret;
	s >> ret;
	if (s)
	  n = ret;
	else
	  n = checknan();
	return s;
      }


      template <typename B> checknan& operator += (const B& b)
      {
	check();
          data += ::goopax::impl::unrange(b);
	TA.write_access();
	return *this;
      }
  
      template <typename B> checknan& operator -= (const B& b)
      {
	check();
	data -= goopax::impl::unrange(b);
	TA.write_access();
	return *this;
      }
  
      template <typename B> checknan& operator *= (const B& b)
      {
	check();
	data *= goopax::impl::unrange(b);
	TA.write_access();
	return *this;
      }
  
      template <typename B> checknan& operator /= (const B& b)
      {
	check();
          data /= ::goopax::impl::unrange(b);
	TA.write_access();
	return *this;
      }
      template <typename B, typename=typename enable_if<is_integral<B>::value>::type> checknan& operator %= (const B& b)
      {
	check();
	data %= goopax::impl::unrange(b);
	TA.write_access();
	return *this;
      }
  
    };






    template <typename T, typename TA> inline T& unrange(checknan<T, TA>& a)
    {
        return a.unrange();
    }


      template <typename A, typename B> struct unrange_opcheck :
	public integral_constant<bool,
				 (!is_same<A, typename unrangetype<A>::type>::value ||
				  !is_same<B, typename unrangetype<B>::type>::value) &&
				   !is_base_of<istream, A>::value &&
				   !is_base_of<ostream, A>::value>
      {};
					
#define RANGE_DEF_OP(OP)						\
      template <typename A, typename B, typename=typename enable_if<unrange_opcheck<A, B>::value>::type>	\
      auto operator OP (const A& a, const B& b)				\
	-> typename remove_reference<decltype(unrange(a) OP unrange(b))>::type \
      {									\
	return unrange(a) OP unrange(b);				\
      }									\
      
#define RANGE_DEF_FUNC(FUNC)						\
      template <typename A, typename B, typename=typename enable_if<unrange_opcheck<A, B>::value>::type> \
      auto FUNC(const A& a, const B& b)					\
	-> typename remove_reference<decltype(FUNC(unrange(a), unrange(b)))>::type \
      {									\
	return FUNC(unrange(a), unrange(b));				\
      }

      RANGE_DEF_FUNC(max)
      RANGE_DEF_FUNC(min)

#undef RANGE_DEF_OP
#undef RANGE_DEF_FUNC

      template <typename T, typename TA> 
      inline decltype(~T()) operator ~ (const checknan<T, TA>& a)
      {
	return ~unrange(a);
      }

#define MAKEFUNC(FUNC) template <typename T, typename TA> inline T FUNC(checknan<T, TA> a) \
      {									\
	return FUNC((typename conditional<is_same<T, half>::value, float, T>::type)unrange(a)); \
      }

      MAKEFUNC(abs)
      MAKEFUNC(floor)
      MAKEFUNC(ceil)
      MAKEFUNC(round)
      MAKEFUNC(isfinite)
      MAKEFUNC(isnormal)
      MAKEFUNC(isnan)
      MAKEFUNC(isinf)

      GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)

      
#undef MAKEFUNC


    namespace range_types
    {
      using Long_double = long double;
      using Double = double;
      using Float = float;
      using Half = half;
      using Bool = bool;
      using Char = char;
      using Schar = signed char;
      using Uchar = unsigned char;
      using Short = short;
      using Int = int;
      using Long = long;
      using Longlong = long long;
      using Ushort = unsigned short;
      using Uint = unsigned int;
      using Ulong = unsigned long;
      using Ulonglong = unsigned long long;
      typedef int8_t Int8_t;
      typedef int16_t Int16_t;
      typedef int32_t Int32_t;
      typedef int64_t Int64_t;
      typedef uint8_t Uint8_t;
      typedef uint16_t Uint16_t;
      typedef uint32_t Uint32_t;
      typedef uint64_t Uint64_t;
      typedef size_t Size_t;

      using Ssize_t = typename make_signed<Size_t>::type;
    }
    using namespace range_types;



    namespace checkrange_types
    {
      namespace debug
      {
	template <typename T> using checknantype = checknan<typename impl::unrangetype<typename remove_const<typename remove_reference<T>::type>::type>::type, impl::thread_access_none>;
	template <typename T> struct checknantypestr
	{
	  typedef checknantype<T> type;
	};

	using Tlong_double = checknantype<long double>;
	using Tdouble = checknantype<double>;
	using Tfloat = checknantype<float>;
	using Thalf = checknantype<half>;
	using Tbool = checknantype<bool>;
	using Tchar = checknantype<char>;
	using Tschar = checknantype<signed char>;
	using Tuchar = checknantype<unsigned char>;
	using Tshort = checknantype<short>;
	using Tint = checknantype<int>;
	using Tlong = checknantype<long>;
	using Tlonglong = checknantype<long long>;
	using Tushort = checknantype<unsigned short>;
	using Tuint = checknantype<unsigned int>;
	using Tulong = checknantype<unsigned long>;
	using Tulonglong = checknantype<unsigned long long>;
	using Tint8_t = checknantype<int8_t>;
	using Tint16_t = checknantype<int16_t>;
	using Tint32_t = checknantype<int32_t>;
	using Tint64_t = checknantype<int64_t>;
	using Tuint8_t = checknantype<uint8_t>;
	using Tuint16_t = checknantype<uint16_t>;
	using Tuint32_t = checknantype<uint32_t>;
	using Tuint64_t = checknantype<uint64_t>;
	using Tsize_t = checknantype<size_t>;
	using Tssize_t = checknantype<Ssize_t>;
	using Tunsigned_int = checknantype<unsigned int>;
      }
      namespace release
      {
	template <typename T> using checknantype = typename impl::unrangetype<typename remove_const<typename remove_reference<T>::type>::type>::type;
	template <typename T> struct checknantypestr
	{
	  typedef checknantype<T> type;
	};

	using Tlong_double = long double;
	using Tdouble = double;
	using Tfloat = float;
	using Thalf = half;
	using Tbool = bool;
	using Tchar = char;
	using Tschar = signed char;
	using Tuchar = unsigned char;
	using Tshort = short;
	using Tint = int;
	using Tlong = long;
	using Tlonglong = long long;
	using Tushort = unsigned short;
	using Tuint = unsigned int;
	using Tulong = unsigned long;
	using Tulonglong = unsigned long long;
	using Tint8_t = int8_t;
	using Tint16_t = int16_t;
	using Tint32_t = int32_t;
	using Tint64_t = int64_t;
	using Tuint8_t = uint8_t;
	using Tuint16_t = uint16_t;
	using Tuint32_t = uint32_t;
	using Tuint64_t = uint64_t;
	using Tsize_t = size_t;
	using Tssize_t = Ssize_t;
	using Tunsigned_int = unsigned int;
      }
    }


    template <typename T> inline bool is_uninitialized(const T&)
    {
      return false;
    }
    template <typename T, typename TA> inline bool is_uninitialized(const checknan<T, TA>& a)
    {
      return a.is_uninitialized();
    }
    template <typename T> inline bool is_initialized(const T&)
    {
      return true;
    }
    template <typename T, typename TA> inline bool is_initialized(const checknan<T, TA>& a)
    {
      return a.is_initialized();
    }

    template <typename T> struct is_half : is_same<typename unrangetype_core<T>::type, Half>
    {};
  }

  using impl::checknan;

  using impl::unrange;
  using namespace impl::range_types;
  using impl::unrangetype;
  using impl::unrangetype_core;
  using impl::is_checknan;
  using impl::is_initialized;
  using impl::is_uninitialized;

  using impl::output_goopax_struct;
    using impl::is_half;
}


namespace std
{
    using goopax::impl::isnan;
    using goopax::impl::isinf;
    using goopax::impl::isfinite;
    using goopax::impl::isnormal;
    using goopax::impl::floor;
    using goopax::impl::ceil;
    using goopax::impl::round;
    using goopax::impl::min;
    using goopax::impl::max;
    using goopax::impl::abs;

    
  template<typename T, typename TA>
  struct numeric_limits<goopax::checknan<T, TA> > :
    public numeric_limits<T>
  {
  };


  template <typename T, typename TA>
  struct make_signed<goopax::checknan<T, TA> >
  {
    using type = goopax::checknan<typename make_signed<T>::type, TA>;
  };

  template <typename T, typename TA>
  struct make_unsigned<goopax::checknan<T, TA> >
  {
    using type = goopax::checknan<typename make_unsigned<T>::type, TA>;
  };
}

namespace goopax
{
  namespace impl
  {
    using goopax::impl::mutex;
    using goopax::impl::lock_guard;
    using goopax::impl::unique_lock;
  }

  namespace release
  {
    namespace types
    {
      using namespace goopax::impl::checkrange_types::release;
    }
  }
  namespace debug
  {
    namespace types
    {
      using namespace goopax::impl::checkrange_types::debug;
    }
  }

  namespace impl
  {
#if GOOPAX_DEBUG
    using namespace goopax::debug::types;
#else
    using namespace goopax::release::types;
#endif
  }
}


namespace goopax
{
  namespace impl
  {
#if GOOPAX_DEBUG
    using namespace goopax::debug::types;
#else
    using namespace goopax::release::types;
#endif
  }
}
