// -*- C++ -*-

#pragma once

#include "./defs.h"
#include <array>
#include <cassert>

#if GOOPAX_HAVE_DEBUG_STL
#include <debug/vector>
#include <debug/array>
#include <debug/string>
#endif
#include <string>
#include <vector>

#include <tuple>


namespace goopax
{
  namespace impl
  {
    template <typename T, typename THREAD_ACCESS = struct thread_access_none> class checknan;
  }
}

#define HALF_ROUND_STYLE 1 // round to nearest
#define HALF_ROUND_TIES_TO_EVEN	1


#include "half/half.hpp"

namespace goopax
{
    namespace impl
    {
      using half_float::isnan;
      using half_float::isinf;
      using half_float::isfinite;
      using half_float::isnormal;
      using namespace half_float;
    }
}



#include <limits>

#include <memory>
#include <cmath>


namespace goopax
{
  namespace impl
  {
    using ::half_float::half;

    using std::pair;
    using std::make_pair;
    using std::abs;
    using std::min;
    using std::max;

    using std::unique_ptr;
    using std::shared_ptr;
    using std::weak_ptr;

    using std::numeric_limits;

    using std::swap;
    using std::ostream;
    using std::istream;
    using std::endl;
    using std::hex;
    using std::flush;

    using std::stringstream;
    using std::istringstream;
    using std::ostringstream;
    using std::streambuf;

    using std::true_type;
    using std::false_type;
    using std::integral_constant;
    using std::declval;


    template <typename T> struct is_floating_point : integral_constant<bool, std::is_floating_point<T>::value || std::is_same<T, half>::value> {};
    using std::is_fundamental;
    template <typename T> struct is_integral : public std::is_integral<T> {};
    template <typename T> struct is_signed : integral_constant<bool, std::is_signed<T>::value || std::is_same<T, half>::value> {};
    template <typename T> struct is_unsigned : public std::is_unsigned<T> {};
    template <typename T> struct is_arithmetic : integral_constant<bool, std::is_arithmetic<T>::value || std::is_same<T, half>::value> {};

    template <typename T, typename TA> struct is_floating_point<checknan<T, TA>> :
      public is_floating_point<T>
    {};
    template <typename T, typename TA> struct is_integral<checknan<T, TA>> :
      public is_integral<T>
    {};
    template <typename T, typename TA> struct is_signed<checknan<T, TA>> :
      public is_signed<T>
    {};
    template <typename T, typename TA> struct is_unsigned<checknan<T, TA>> :
      public is_unsigned<T>
    {};
    template <typename T, typename TA> struct is_arithmetic<checknan<T, TA>> :
      public is_arithmetic<T>
    {};


    
    using std::is_pointer;
    using std::conditional;
    using std::is_base_of;
    using std::is_same;
    using std::is_class;
    using std::is_scalar;
    using std::enable_if;
    using std::make_signed;
    using std::make_unsigned;
      
    using std::remove_pointer;
    using std::remove_reference;
    using std::remove_const;
    using std::remove_cv;

    using std::type_info;
    using std::initializer_list;
    using std::ios;
    using std::ios_base;

    using std::move;

    using std::tuple;
    using std::tuple_size;
#if GOOPAX_USE_DEBUG_STL
    using std::__debug::vector;
    using std::__debug::array;
    using __gnu_debug::basic_string;
    using __gnu_debug::string;
#else
    using std::vector;
    using std::array;
    using std::string;
#endif


    
    /*
#if GOOPAX_DEBUG
    using namespace goopax::debug;
#else
    using namespace goopax::release;
#endif
    using namespace types;
    */
    
    
    template <class T> struct isplain :
      public integral_constant<bool, (is_fundamental<T>::value ||
				      is_same<T, half>::value ||
				      is_pointer<T>::value)>
    {};

    template <class A, class B> struct isplain<pair<A,B> > :
      public integral_constant<bool, isplain<A>::value && isplain<B>::value>
    {};

  }
}


